#include "assembler.h"
#include <unistd.h>
#include <stdlib.h>
#include <strings.h>
#include <stdio.h>

/* Define it as 1 to see what happens during the execution */
#define VERBOSE 0

#define RAM_SIZE (10*1024)
#define MAX_CODE_SIZE (512)
#define MAX_DATA_SIZE (32)
#define MAX_STACK_SIZE (32)

static unsigned char memory[RAM_SIZE];
static unsigned int load_address;
static addr_t current_instruction;
static addr_t stack_pointer;
static unsigned int end;
static unsigned char exit_status;

void doassign(instruction_t ins) {
  addr_t dest, src;
  unsigned char src_value;
#if VERBOSE==1
  printf("Doing assign ");
#endif
  dest = ins.datas.assign.arg1.addr;
  switch(GET_MODE(ins.datas.assign.arg1.type)) {
  case CONSTANT: fprintf(stderr,"Affectation d'une constante ?\n"); exit(1);
  case INDIRECT: dest = (memory+load_address+MAX_CODE_SIZE)[dest];
  }
  src = ins.datas.assign.arg2.addr;
  switch(GET_MODE(ins.datas.assign.arg2.type)) {
  case CONSTANT: src_value = ins.datas.assign.arg2.addr; break;
  case INDIRECT: src = (memory+load_address+MAX_CODE_SIZE)[src];
  case DIRECT: src_value = (memory+load_address+MAX_CODE_SIZE)[src];
  }
  /* Now do the trick! */
  switch(ins.opcode) {
  case ASSIGN_PLUS: (memory+load_address+MAX_CODE_SIZE)[dest]  += src_value;
#if VERBOSE==1
                    printf("M[%d] += %d\n",dest,src_value);
#endif
                    break;
  case ASSIGN_MINUS: (memory+load_address+MAX_CODE_SIZE)[dest] -= src_value;
#if VERBOSE==1
                    printf("M[%d] -= %d\n",dest,src_value);
#endif
                    break;
  case ASSIGN_MULT: (memory+load_address+MAX_CODE_SIZE)[dest]  *= src_value;
#if VERBOSE==1
                    printf("M[%d] *= %d\n",dest,src_value);
#endif
                    break;
  case ASSIGN:      (memory+load_address+MAX_CODE_SIZE)[dest]   = src_value;
#if VERBOSE==1
                    printf("M[%d]  = %d\n",dest,src_value);
#endif
                    break;
  }
}

void doif(instruction_t ins) {
  unsigned char src1_value, src2_value;
  addr_t src1, src2;

#if VERBOSE==1
  printf("Doing an if ");
#endif
  src1 = ins.datas.si.arg1.addr;
  switch(GET_MODE(ins.datas.si.arg1.type)) {
  case CONSTANT: src1_value = ins.datas.si.arg1.addr; break;
  case INDIRECT: src1 = (memory+load_address+MAX_CODE_SIZE)[src1];
  case DIRECT: src1_value = (memory+load_address+MAX_CODE_SIZE)[src1];
  }
  src2 = ins.datas.si.arg2.addr;
  switch(GET_MODE(ins.datas.si.arg2.type)) {
  case CONSTANT: src2_value = ins.datas.si.arg2.addr; break;
  case INDIRECT: src2 = (memory+load_address+MAX_CODE_SIZE)[src2];
  case DIRECT: src2_value = (memory+load_address+MAX_CODE_SIZE)[src2];
  }
  /* Do the test */
#if VERBOSE==1
  printf("%d =? %d",src1_value,src2_value);
#endif
  if (src1_value==src2_value) {
    /* Transfer the control to the label! */
    current_instruction = ins.datas.si.arg3.addr;
#if VERBOSE==1
    printf(" yes then jump to %d",current_instruction);
#endif
  }
#if VERBOSE==1
  printf("\n");
#endif
}

void dojump(instruction_t ins) {
#if VERBOSE==1
  printf("Doing a jump ");
#endif
  current_instruction = ins.datas.jump.arg1.addr;
#if VERBOSE==1
  printf("to %d\n",current_instruction);
#endif
}

void docall(instruction_t ins) {
#if VERBOSE==1
  printf("Doing a call to %d\n",ins.datas.call.arg1.addr);
#endif
  if (stack_pointer>MAX_STACK_SIZE) {
    fprintf(stderr,"Depassement de la capacite de la pile!\n");
    exit(1);
  }
  /* Push the return address onto the stack */
  (memory+load_address+MAX_CODE_SIZE+MAX_DATA_SIZE)[stack_pointer++] = current_instruction;
  /* Transfer execution */
  current_instruction = ins.datas.call.arg1.addr;
}

void doret(instruction_t ins) {
#if VERBOSE==1
  printf("Doing a return ");
#endif
  if (stack_pointer==0) { 
    fprintf(stderr,"Probleme avec le pointeur de pile!!!\n");
    exit(1);
  }
  /* Transfer execution to the value pop'ed from the top */
  current_instruction = (memory+load_address+MAX_CODE_SIZE+MAX_DATA_SIZE)[--stack_pointer];
#if VERBOSE==1
  printf("to %d\n",current_instruction);
#endif
}

void dolck(instruction_t ins) {
  switch(ins.opcode) {
  case LCK:
    fprintf(stderr,"LCK not yet implemented\n"); break;
  case UNLCK:
    fprintf(stderr,"UNLCK not yet implemented\n"); break;
  }
}

void dofork(instruction_t ins) {
  fprintf(stderr,"FORK not yet implemented\n");
}

void doexec(instruction_t ins) {
  fprintf(stderr,"EXEC not yet implemented\n");
}

void doexit(instruction_t ins) {
  addr_t addr;
  unsigned char value;

#if VERBOSE==1
  printf("Doing an exit ");
#endif
  addr = ins.datas.exit.arg1.addr;
  switch(GET_MODE(ins.datas.exit.arg1.type)) {
  case CONSTANT: value = ins.datas.exit.arg1.addr; break;
  case INDIRECT: addr = (memory+load_address+MAX_CODE_SIZE)[addr];
  case DIRECT: value = (memory+load_address+MAX_CODE_SIZE)[addr];
  }
  exit_status = value;
#if VERBOSE==1
  printf("with %d\n",exit_status);
#endif
}

void dosignal(instruction_t ins) {
  fprintf(stderr,"SIGNAL not yet implemented\n");
}

void dokill(instruction_t ins) {
  fprintf(stderr,"KILL not yet implemented\n");
}

void dogetppid(instruction_t ins) {
  fprintf(stderr,"GETPPID not yet implemented\n");
}

void doread(instruction_t ins) {
  fprintf(stderr,"READ not yet implemented\n");
}

void dowrite(instruction_t ins) {
  unsigned char src_value;
  addr_t src;
#if VERBOSE==1
  printf("Doing a write\n");
#endif
  src = ins.datas.assign.arg1.addr;
  switch(GET_MODE(ins.datas.assign.arg1.type)) {
  case CONSTANT: src_value = ins.datas.assign.arg1.addr; break;
  case INDIRECT: src = (memory+load_address+MAX_CODE_SIZE)[src];
  case DIRECT: src_value = (memory+load_address+MAX_CODE_SIZE)[src];
  }
#if VERBOSE==1
  printf("OUTPUT %c\n",src_value);
#else
  putchar(src_value);
#endif
}

void dowait(instruction_t ins) {
  fprintf(stderr,"WAIT not yet implemented\n");
}

int main() {
  int           i = 0;
  instruction_t ins;

  load_address = 100; /* A program can be loaded anywhere */

  i = 0;
  while ( read(0,&ins,sizeof(instruction_t))==sizeof(instruction_t) ) {
    *((instruction_t *)(memory+load_address+i*sizeof(instruction_t))) = ins;
    i++;
  }
  printf("Program loaded. program length %d instructions\n",i);
  end = 0;
  current_instruction = 0;
  stack_pointer = 0;
  exit_status = 0;
  while (!end) {
    /* Fetch the current instruction */
    ins = *((instruction_t *)(memory+load_address+current_instruction*sizeof(instruction_t)));
    /* Then point to the next instruction */
    current_instruction++;
    /* Now execute the instruction */
    switch(ins.opcode) {
    case ASSIGN_PLUS:
    case ASSIGN_MINUS:
    case ASSIGN_MULT:
    case ASSIGN:       doassign(ins); break;
    case IF:           doif(ins); break;
    case JUMP:         dojump(ins); break;
    case CALL:         docall(ins); break;
    case RET:          doret(ins); break;
    case UNLCK:
    case LCK:          dolck(ins); break;
    case FORK:         dofork(ins); break;
    case EXEC:         doexec(ins); break;
    case EXIT:         doexit(ins); end = 1; break;
    case SIGNAL:       dosignal(ins); break;
    case WAIT:         dowait(ins); break;
    case KILL:         dokill(ins); break;
    case READ:         doread(ins); break;
    case WRITE:        dowrite(ins); break;
    case GETPPID:      dogetppid(ins); break;
    }
  }
  printf("THE END\n");
  return 0;
}
